#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>

#define IPT_LOOPREDIRECT_OPT_DST      0x01
#define IPT_LOOPREDIRECT_OPT_SRC      0x02

#if defined(IPTABLES_VERSION)
#define COMPATIBLE_VERSION IPTABLES_VERSION
#else
#define COMPATIBLE_VERSION NETFILTER_VERSION
#endif

struct ip_nat_multi_range_pair{
  struct ip_nat_multi_range src,dst;
};

/* Dest NAT data consists of a multi-range, indicating where to map
   to. */
struct ipt_natinfo
{
  struct ipt_entry_target t;
  struct ip_nat_multi_range mr;
};

/* Function which prints out usage message. */
static void
help(void)
{
  printf(
	 "LOOPREDIRECT v%s options:\n"
	 " --to-dst <ipaddr>[-<ipaddr>]:<port>[-<port>]\n"
	 "				Address to map dest to.\n"
	 " --src-ip <ipaddr>\n"
	 "                              Faked source IP.\n\n",
	 COMPATIBLE_VERSION);
}

static struct option opts[] = {
  { "to-dst", 1, 0, '1' },
  { "src-ip", 1, 0, '2' },
  { 0 }
};

/* Initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache)
{
  struct ip_nat_multi_range_pair *mrp = (struct ip_nat_multi_range_pair *)t->data;

  /* Actually, it's 0, but it's ignored at the moment. */
  mrp->src.rangesize=1;
  mrp->dst.rangesize=1;

  /* Can't cache this */
  *nfcache |= NFC_UNKNOWN;
}

/* Ranges expected in network order. */
static void
parse_to(char *arg, int portok, struct ip_nat_multi_range *mr)
{
  char *colon, *dash;
  struct in_addr *ip;

  memset(mr, 0, sizeof(*mr));
  colon = strchr(arg, ':');

  if (colon) {
    int port;

    if (!portok)
      exit_error(PARAMETER_PROBLEM,
		 "Need TCP or UDP with port specification");

    mr->range[0].flags |= IP_NAT_RANGE_PROTO_SPECIFIED;

    port = atoi(colon+1);
    if (port == 0 || port > 65535)
      exit_error(PARAMETER_PROBLEM,
		 "Port `%s' not valid\n", colon+1);

    dash = strchr(colon, '-');
    if (!dash) {
      mr->range[0].min.tcp.port
	= mr->range[0].max.tcp.port
	= htons(port);
    } else {
      int maxport;

      maxport = atoi(dash + 1);
      if (maxport == 0 || maxport > 65535)
	exit_error(PARAMETER_PROBLEM,
		   "Port `%s' not valid\n", dash+1);
      if (maxport < port)
	/* People are stupid. */
	exit_error(PARAMETER_PROBLEM,
		   "Port range `%s' funky\n", colon+1);
      mr->range[0].min.tcp.port = htons(port);
      mr->range[0].max.tcp.port = htons(maxport);
    }
    /* Starts with a colon? No IP info...*/
    if (colon == arg)
      exit_error(PARAMETER_PROBLEM,
		 "No dst IP\n");
    *colon = '\0';
  }

  mr->range[0].flags |= IP_NAT_RANGE_MAP_IPS;
  dash = strchr(arg, '-');
  if (colon && dash && dash > colon)
    dash = NULL;

  if (dash)
    *dash = '\0';

  ip = dotted_to_addr(arg);
  if (!ip)
    exit_error(PARAMETER_PROBLEM, "Bad IP address `%s'\n",
	       arg);
  mr->range[0].min_ip = ip->s_addr;
  if (dash) {
    ip = dotted_to_addr(dash+1);
    if (!ip)
      exit_error(PARAMETER_PROBLEM, "Bad IP address `%s'\n",
		 dash+1);
    mr->range[0].max_ip = ip->s_addr;
  } else
    mr->range[0].max_ip = mr->range[0].min_ip;

  mr->rangesize=1;
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target)
{
  struct ip_nat_multi_range_pair *mrp
    = (struct ip_nat_multi_range_pair *)(*target)->data;
  int portok;

  if (entry->ip.proto == IPPROTO_TCP
      || entry->ip.proto == IPPROTO_UDP)
    portok = 1;
  else
    portok = 0;

  switch (c) {
  case '1':
    if (!portok)
      exit_error(PARAMETER_PROBLEM,
		 "Need TCP or UDP with port specification");

    if (check_inverse(optarg, &invert, NULL, 0))
      exit_error(PARAMETER_PROBLEM,
		 "Unexpected `!' after --to-dst");

    parse_to(optarg, portok, &mrp->dst);
    
    *flags|=IPT_LOOPREDIRECT_OPT_DST;
    return 1;

  case '2':
    if (!portok)
      exit_error(PARAMETER_PROBLEM,
		 "Need TCP or UDP with port specification");

    if (check_inverse(optarg, &invert, NULL, 0))
      exit_error(PARAMETER_PROBLEM,
		 "Unexpected `!' after --src-ip");

    mrp->src.range[0].min_ip=mrp->src.range[0].max_ip=(dotted_to_addr(optarg))->s_addr;
    *flags|=IPT_LOOPREDIRECT_OPT_SRC;
    return 1;

  default:
    return 0;
  }
}

/* Final check; both --to-dst and --src-ip must be set. */
static void final_check(unsigned int flags)
{
  if(!(flags & IPT_LOOPREDIRECT_OPT_DST) || !(flags & IPT_LOOPREDIRECT_OPT_SRC))
    exit_error(PARAMETER_PROBLEM,"LOOPREDIRECT target: --to-dst and --src-ip must be set.");
}

static void print_range(const struct ip_nat_range *r)
{
  if (r->flags & IP_NAT_RANGE_MAP_IPS) {
    struct in_addr a;

    a.s_addr = r->min_ip;
    printf("%s", addr_to_dotted(&a));
    if (r->max_ip != r->min_ip) {
      a.s_addr = r->max_ip;
      printf("-%s", addr_to_dotted(&a));
    }
  }
  if (r->flags & IP_NAT_RANGE_PROTO_SPECIFIED) {
    printf(":");
    printf("%hu", ntohs(r->min.tcp.port));
    if (r->max.tcp.port != r->min.tcp.port)
      printf("-%hu", ntohs(r->max.tcp.port));
  }
}

/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target,
      int numeric)
{
  struct ip_nat_multi_range_pair *mrp
    = (struct ip_nat_multi_range_pair *)target->data;
  unsigned int i=0;

  printf("to dest:");
  for(i=0; i<mrp->dst.rangesize; i++){
    print_range(&mrp->dst.range[i]);
    printf(" ");
  }

  printf("fake source:");
  printf("%s", addr_to_dotted((void *)&mrp->src.range[0].min_ip));
}

/* Saves the union ipt_targinfo in parsable form to stdout. */
static void
save(const struct ipt_ip *ip, const struct ipt_entry_target *target)
{
  struct ip_nat_multi_range_pair *mrp
    = (struct ip_nat_multi_range_pair *)target->data;
  unsigned int i=0;

  for(i=0; i<mrp->dst.rangesize; i++){
    printf("--to-dst ");
    print_range(&mrp->dst.range[i]);
    printf(" ");
  }

  printf("--fake-src ");
  printf("%s", addr_to_dotted((void *)&mrp->src.range[0].min_ip));
}

static
struct iptables_target loopredir
= { NULL,
    "LOOPREDIRECT",
    NETFILTER_VERSION,
    IPT_ALIGN(sizeof(struct ip_nat_multi_range_pair)),
    IPT_ALIGN(sizeof(struct ip_nat_multi_range_pair)),
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    &save,
    opts
};

void _init(void)
{
  register_target(&loopredir);
}

